
from .utils import create_directories_from_paths
from .utils import print_and_convert_size
from .utils import get_or_print_filesize
from .utils import deep_getsizeof
from .utils import get_filesize


def save_filepath_config(FPATHS, overwrite=True, output_fpath='config/filepaths.json',
                        verbose=False):
    """
    Save the filepaths to a JSON file.

    Parameters:
    FPATHS (dict): A dictionary containing the filepaths.
    overwrite (bool): Whether to overwrite the existing file if it already exists. Default is True.
    output_fpath (str): The output filepath to save the JSON file. Default is 'config/filepaths.json'.
    verbose (bool): Whether to print the saved filepaths. Default is False.

    Returns:
    dict: The dictionary containing the filepaths.

    Raises:
    Exception: If the output file already exists and overwrite is set to False.
    """
    import os
    import json
    from pprint import pprint
    
    os.makedirs(os.path.dirname(output_fpath), exist_ok=True)
    if os.path.exists(output_fpath) and not overwrite:
        raise Exception(f"- {output_fpath} already exists and overwrite is set to False.")

    with open(output_fpath, 'w') as f:
        json.dump(FPATHS, f)

    print(f"[i] Filepath json saved as {output_fpath} ")
    if verbose:
        pprint(FPATHS)
        
    return FPATHS

import os





def load_filepath_config(config_fpath = 'config/filepaths.json', verbose=True):
    """
    Loads the filepaths configuration from a JSON file.

    Parameters:
    - config_fpath (str): The filepath of the JSON configuration file. Default is 'config/filepaths.json'.
    - verbose (bool): Whether to print the loaded filepaths. Default is True.

    Returns:
    - dict: A dictionary containing the loaded filepaths.
    """
    import json 
    from pprint import pprint
    with open (config_fpath) as f:
        FPATHS = json.load(f)
    
    if verbose == True:
        print(f"- Filepaths loaded successfully:")
        pprint(FPATHS)
        
    return FPATHS





import statsmodels.tsa.api as tsa
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, mean_absolute_percentage_error

def get_adfuller_results(ts, alpha=.05, label='adfuller', **kwargs):
    """
    Uses statsmodels' adfuller function to test a univariate time series for stationarity.
    Null hypothesis: The time series is NOT stationary. (It "has a unit root".)
    Interpretation: A p-value less than alpha (.05) means the ts IS stationary.
    (We reject the null hypothesis that it is not stationary.)

    Parameters:
    - ts (array-like): The time series data.
    - alpha (float): The significance level used for interpreting p-value. Default is 0.05.
    - label (str): The label for the output DataFrame. Default is 'adfuller'.
    - **kwargs: Additional keyword arguments to be passed to statsmodels' adfuller function.

    Returns:
    - results (DataFrame): DataFrame with the following columns/results:
        - "Test Statistic": The adfuller test statistic.
        - "# of Lags Used": The number of lags used in the calculation.
        - "# of Observations": The number of observations used.
        - "p-value": The p-value for the hypothesis test.
        - "alpha": The significance level used for interpreting p-value.
        - "sig/stationary?": A simplified interpretation of the p-value.

    ADFULLER DOCUMENTATION:
    For the full adfuller documentation, see:
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.stattools.adfuller.html
    """
    (test_stat, pval, nlags, nobs, crit_vals_d, icbest) = tsa.adfuller(ts, **kwargs)
    adfuller_results = {
        'Test Statistic': test_stat,
        '# of Lags Used': nlags,
        '# of Observations': nobs,
        'p-value': round(pval, 6),
        'alpha': alpha,
        'sig/stationary?': pval < alpha
    }
    return pd.DataFrame(adfuller_results, index=[label])


def get_sig_lags(ts, type='ACF', nlags=None, alpha=0.5):
    """
    Calculates the significant lags for the autocorrelation function (ACF) or partial autocorrelation function (PACF) of a time series.

    Parameters:
    - ts (array-like): The time series data.
    - type (str): The type of lag to calculate. Must be either 'ACF' or 'PACF'. Default is 'ACF'.
    - nlags (int): The number of lags to calculate. Default is None.
    - alpha (float): The significance level for determining significant lags. Default is 0.5.

    Returns:
    - sig_lags (array-like): The significant lags.

    Raises:
    - Exception: If the type is not 'ACF' or 'PACF'.
    """
    if type == 'ACF':
        corr_values, conf_int = tsa.stattools.acf(ts, alpha=alpha, nlags=nlags)
    elif type == 'PACF':
        corr_values, conf_int = tsa.stattools.pacf(ts, alpha=alpha, nlags=nlags)
    else:
        raise Exception("type must be either 'ACF' or 'PACF'")

    lags = range(len(corr_values))
    corr_df = pd.DataFrame({
        type: corr_values,
        'Lags': lags,
        'lower ci': conf_int[:, 0] - corr_values,
        'upper ci': conf_int[:, 1] - corr_values
    })
    corr_df = corr_df.set_index("Lags")

    filter_sig_lags = (corr_df[type] < corr_df['lower ci']) | (corr_df[type] > corr_df['upper ci'])
    sig_lags = corr_df.index[filter_sig_lags]
    sig_lags = sig_lags[sig_lags != 0]

    return sig_lags


def plot_acf_pacf(ts, nlags=40, figsize=(10, 5), annotate_sig=False, alpha=.05, acf_kws={}, pacf_kws={}, annotate_seas=False, m=None, seas_color='black'):
    """
    Plots the autocorrelation function (ACF) and partial autocorrelation function (PACF) of a time series.

    Parameters:
    - ts (array-like): The time series data.
    - nlags (int): The number of lags to include in the plot. Default is 40.
    - figsize (tuple): The size of the figure. Default is (10, 5).
    - annotate_sig (bool): Whether to annotate significant lags. Default is False.
    - alpha (float): The significance level for determining significant lags. Default is 0.05.
    - acf_kws (dict): Additional keyword arguments to be passed to statsmodels' plot_acf function.
    - pacf_kws (dict): Additional keyword arguments to be passed to statsmodels' plot_pacf function.
    - annotate_seas (bool): Whether to annotate seasonal lags. Default is False.
    - m (int): The number of observations per season. Required if annotate_seas is True.
    - seas_color (str): The color of the seasonal lines. Default is 'black'.

    Returns:
    - fig (Figure): The matplotlib Figure object.
    """
    fig, axes = plt.subplots(nrows=2, figsize=figsize)
    sig_vline_kwargs = dict(ls=':', lw=1, zorder=0, color='red')

    tsa.graphics.plot_acf(ts, ax=axes[0], lags=nlags, **acf_kws)

    if annotate_sig:
        sig_acf_lags = get_sig_lags(ts, nlags=nlags, alpha=alpha, type='ACF')
        for lag in sig_acf_lags:
            axes[0].axvline(lag, label='sig', **sig_vline_kwargs)

    tsa.graphics.plot_pacf(ts, ax=axes[1], lags=nlags, **pacf_kws)

    if annotate_sig:
        sig_pacf_lags = get_sig_lags(ts, nlags=nlags, alpha=alpha, type='PACF')
        for lag in sig_pacf_lags:
            axes[1].axvline(lag, label='sig', **sig_vline_kwargs)

    if annotate_seas:
        if m is None:
            raise Exception("Must define value of m if annotate_seas=True.")

        n_seasons = nlags // m
        seas_vline_kwargs = dict(ls='--', lw=1, alpha=.7, color=seas_color, zorder=-1)

        for i in range(1, n_seasons + 1):
            axes[0].axvline(m * i, **seas_vline_kwargs, label="season")
            axes[1].axvline(m * i, **seas_vline_kwargs, label="season")

    fig.tight_layout()
    return fig


def plot_forecast(ts_train, ts_test, forecast_df, n_train_lags=None, 
                  figsize=(10,4), title='Comparing Forecast vs. True Data'):
    """
    Plots the training data, test data, and forecast with confidence intervals.

    Parameters:
    ts_train (pandas.Series): Time series data for training.
    ts_test (pandas.Series): Time series data for testing.
    forecast_df (pandas.DataFrame): DataFrame containing forecasted values and confidence intervals.
    n_train_lags (int, optional): Number of training lags to plot. If not specified, all training data will be plotted.
    figsize (tuple, optional): Figure size. Default is (10, 4).
    title (str, optional): Title of the plot. Default is 'Comparing Forecast vs. True Data'.

    Returns:
    fig (matplotlib.figure.Figure): The generated figure.
    ax (matplotlib.axes.Axes): The axes of the generated figure.
    """
    ### PLot training data, and forecast (with upper/,lower ci)
    fig, ax = plt.subplots(figsize=figsize)

    # setting the number of train lags to plot if not specified
    if n_train_lags==None:
        n_train_lags = len(ts_train)
            
    # Plotting Training  and test data
    ts_train.iloc[-n_train_lags:].plot(ax=ax, label="train")
    ts_test.plot(label="test", ax=ax)

    # Plot forecast
    forecast_df['mean'].plot(ax=ax, color='green', label="forecast")

    # Add the shaded confidence interval
    ax.fill_between(forecast_df.index, 
                    forecast_df['mean_ci_lower'],
                   forecast_df['mean_ci_upper'],
                   color='green', alpha=0.3,  lw=2)

    # set the title and add legend
    ax.set_title(title)
    ax.legend();
    
    return fig, ax


def thiels_U(ts_true, ts_pred):
    """Calculate's Thiel's U metric for forecasting accuracy.
    Accepts true values and predicted values.
    Original Formula Source: https://docs.oracle.com/cd/E57185_01/CBREG/ch06s02s03s04.html
    Adapted Function from Source: https://github.com/jirvingphd/predicting-the-SP500-using-trumps-tweets_capstone-project/blob/cf11f6ed88721433d2c00cb1f8486206ab179cc0/bsds/my_keras_functions.py#L735
    Returns: 
        U (float)
        
    Thiel's U Value Interpretation:
    - <1  = Forecasting is better than guessing 
    - 1   = Forecasting is about as good as guessing
    - >1  = Forecasting is worse than guessing 
    """
    import numpy as np
    # sum_list = []
    num_list=[]
    denom_list=[]
    
    for t in range(len(ts_true)-1):
        
        num_exp = (ts_pred[t+1] - ts_true[t+1])/ts_true[t]
        num_list.append([num_exp**2])
        
        denom_exp = (ts_true[t+1] - ts_true[t])/ts_true[t]
        denom_list.append([denom_exp**2])
        
    U = np.sqrt( np.sum(num_list) / np.sum(denom_list))
    return U


def regression_metrics_ts(ts_true, ts_pred, label="", verbose=True, output_dict=False,
                          thiels_U=False):
    """
    Calculates regression metrics for comparing true and predicted time series data.

    Parameters:
    - ts_true (array-like): The true time series data.
    - ts_pred (array-like): The predicted time series data.
    - label (str): The label for the metrics. Default is an empty string.
    - verbose (bool): Whether to print the metrics. Default is True.
    - output_dict (bool): Whether to return the metrics as a dictionary. Default is False.
    - thiels_U (bool): Whether to calculate Thiel's U metric. Default is False.
                        - See docstring for thiels_U function for interpretation.

    Returns:
    - metrics (dict): The regression metrics as a dictionary. Only returned if output_dict is True.
    """
    mae = mean_absolute_error(ts_true, ts_pred)
    mse = mean_squared_error(ts_true, ts_pred)
    rmse = mean_squared_error(ts_true, ts_pred, squared=False)
    r_squared = r2_score(ts_true, ts_pred)
    mae_perc = mean_absolute_percentage_error(ts_true, ts_pred) * 100
    if thiels_U:
        U = thiels_U(ts_true, ts_pred)
    if verbose:
        header = "---" * 20
        print(header, f"Regression Metrics: {label}", header, sep="\n")
        print(f"- MAE = {mae:,.3f}")
        print(f"- MSE = {mse:,.3f}")
        print(f"- RMSE = {rmse:,.3f}")
        print(f"- R^2 = {r_squared:,.3f}")
        print(f"- MAPE = {mae_perc:,.2f}%")
        
        if thiels_U:
            print(f"- Thiel's U = {U:,.2f}")
            
    if output_dict:
        metrics = {
            "Label": label,
            "MAE": mae,
            "MSE": mse,
            "RMSE": rmse,
            "R^2": r_squared,
            "MAPE(%)": mae_perc,
        }
        if thiels_U:
            metrics['Thiel\'s U'] = U
        return metrics





def plot_acf(ts, nlags=None, alpha=0.05, figsize=(12,4), annotate_sig=False): 
    """
    Plots the autocorrelation function (ACF) for a given time series.

    Parameters:
    - ts: The time series data.
    - nlags: The number of lags to include in the plot. If None, all lags are included.
    - alpha: The significance level for determining significant lags.
    - figsize: The size of the figure (width, height) in inches.
    - annotate_sig: Whether to annotate significant lags with red dashed lines.

    Returns:
    - None
    """
    fig, ax = plt.subplots(figsize=figsize) 
    tsa.graphics.plot_acf(ts, ax=ax, lags=nlags, alpha=alpha)

    if annotate_sig == True:
        sig_lags = get_sig_lags(ts,nlags=nlags,alpha=alpha, type='ACF')
        
        ## ADDING ANNOTATING SIG LAGS
        for lag in sig_lags:
            ax.axvline(lag, ls='--', lw=1, zorder=0, color='red')
            
            

def plot_pacf(ts, nlags=None, alpha=0.05, figsize=(12,4), annotate_sig=False): 
    """
    Plots the partial autocorrelation function (PACF) for a given time series.

    Parameters:
    - ts: The time series data.
    - nlags: The number of lags to include in the plot. If None, all lags will be included.
    - alpha: The significance level for determining significant lags.
    - figsize: The size of the figure (width, height) in inches.
    - annotate_sig: Whether to annotate significant lags with red dashed lines.

    Returns:
    - None
    """
    fig, ax = plt.subplots(figsize=figsize) 
    tsa.graphics.plot_pacf(ts, ax=ax, lags=nlags, alpha=alpha)

    if annotate_sig == True:
        sig_lags = get_sig_lags(ts,nlags=nlags,alpha=alpha, type='PACF')
        
        ## ADDING ANNOTATING SIG LAGS
        for lag in sig_lags:
            ax.axvline(lag, ls='--', lw=1, zorder=0, color='red')


def display_ts_workflow(return_text=False):
    """
    Displays the Time Series Modeling Workflow - (S)ARIMA markdown document.

    Parameters:
    - return_text (bool): If True, returns the text content of the markdown document as a string. Default is False.

    Returns:
    - str: The text content of the markdown document, if return_text is True.

    """
    import requests
    try:
        from IPython.display import Markdown, display
        display_func = lambda x: display(Markdown(x))
    except:
        display_func = print
    url = "https://raw.githubusercontent.com/coding-dojo-data-science/dojo_ds/main/assets/Time%20Series%20Modeling%20Workflow%20-%20(S)ARIMA.md"
    resp = requests.get(url)
    display_func(resp.text)

    if return_text:
        return resp.text









# def reference_interpreting_acf_pacf(seasonal=True, return_string = False):
#     table = """**Determining (S)ARIMA Orders from ACF/PACF**:\n
# |                  | AR(p)                | MA(q)               | ARMA(p,q) |
# |:----- | :---------------: | :---------------:|:---------------:|
# | **ACF**              | Gradually decreases  | Dramatic drop after lag \(q\) | Gradually decreases |
# | **PACF**             | Dramatic drop after lag \(p\) | Gradually decreases      | Gradually decreases  |
# | **ARIMA Order** (p,d,q) | (p,d,0) | (0,d,q) | Start with (1,d,1) & Iterate|

# ___

# **If seasonality is present**

# | ***If seasonal*** | S-AR(P)                | S-MA(Q)                | SARIMA(P,Q)            |
# |:-----------------| :---------------: | :---------------:|:---------------:|  
# | **ACF** (seasonal lags)†         | Gradually decreases   | Dramatic drop after lag \(Q\)| Gradually decreases | 
# | **PACF** (seasonal lags)†       |  Dramatic drop after lag \(P\) | Gradually decreases    | Gradually decreases |
# | **Seasonal Order** (P,D,Q)  | (P,D,0)      | (0,D,Q)       | Start with (1,D,1) & Iterate  | 

# - † seasonal lags = lags that are a multiple of the season length (m). E.g., If daily, m=7, check lags 7,14,21,etc.

# """
#     if return_string == True:
#         return table
#     else: 
#         from IPython.display import Markdown, display
#         display(Markdown(table))


def reference_interpreting_acf_pacf(seasonal=True, return_string=False):
    """
    Generates a table for determining (S)ARIMA orders from ACF/PACF.

    Args:
        seasonal (bool, optional): Indicates if seasonality is present. Defaults to True.
        return_string (bool, optional): Indicates if the table should be returned as a string. 
            If False, the table is displayed using IPython's Markdown. Defaults to False.

    Returns:
        str or None: The table as a string if return_string is True, otherwise None.
    """
    table = """#### **Determining (S)ARIMA Orders from ACF/PACF**:\n
|                  | AR(p)                | MA(q)               | ARMA(p,q) |
|:----- | :---------------: | :---------------:|:---------------:|
| **ACF**              | Gradually decreases  | Dramatic drop after lag \(q\) | Gradually decreases |
| **PACF**             | Dramatic drop after lag \(p\) | Gradually decreases      | Gradually decreases  |
| **ARIMA Order (p,d,q)** | **(p,d,0)** | **(0,d,q)** | **Start with (1,d,1) & Iterate**|


___


#### **Determining Seasonal Orders** (only if seasonality is present):

| ***If seasonal*** | S-AR(P)                | S-MA(Q)                | SARIMA(P,Q)            |
|:-----------------| :---------------: | :---------------:|:---------------:|  
| **ACF** (seasonal lags)†         | Gradually decreases   | Dramatic drop after lag \(Q\)| Gradually decreases | 
| **PACF** (seasonal lags)†       |  Dramatic drop after lag \(P\) | Gradually decreases    | Gradually decreases |
| **Seasonal Order (P,D,Q)**  | **(P,D,0)**      | **(0,D,Q)**       | **Start with (1,D,1) & Iterate**  | 

- † seasonal lags = lags that are a multiple of the season length (m). E.g., If daily, m=7, check lags 7,14,21,etc.

"""
    if seasonal==False:
        table = table.split('___')[0]
        table = table.replace("(S)ARIMA",'ARIMA')
   
    
             
    if return_string:
        return table
    else: 
        try:
            from IPython.display import Markdown, display
            display_func = lambda x: display(Markdown(x))
        except:
            display_func = print
            
        display_func(table)


import matplotlib.pyplot as plt
import seaborn as sns

####### PREVIOUS (slightly updated to only return fig)

def summarize_df(df_):
    """Source: Insights for Stakeholder Lesson - https://login.codingdojo.com/m/0/13079/91969 
    Example Usage:
    >> df = pd.read_csv(filename)
    >> summary = summarize_df(df)
    >> summary"""
    import pandas as pd
    df = df_.copy()
    report = pd.DataFrame({
                        'dtype':df.dtypes,
                        '# null': df.isna().sum(),
                        'null (%)': df.isna().sum()/len(df)*100,
                        'nunique':df.nunique(),
                        "min":df.min(),
                        'max':df.max()
             })
    report.index.name='Column'
    return report.reset_index()
  
  
  
def explore_numeric(df, x, figsize=(6,5), show=True):
  """Plots a Seaborn histplot on the top subplot and a horizontal boxplot on he bottom.
    Additionally, prints information on: 
    - the # and % of null values
    - number of unique values
    - the most frequent value and how often frequent it is (%)
    - A warning message if the feature is quasi-constant or constant feature
                            (if more than 99% of feature is a single value)

    Args:
        df (Frame): DataFrame that contains column x
        x (str): a column name 
        fillna (bool, optional): if True, fillna with the placeholder. Defaults to True.
        placeholder (str, optional): Value used to fillna if fillna is True. Defaults to 'MISSING'.
        figsize (tuple, optional): Figure size (width, height). Defaults to (6,5).
        order (list, optional): List of categories to include in graph, in the specified order. Defaults to None. 
                                Note: any category not in the order list will not be shown on the graph.
                                    If a category is included in the order list that isn't in the data, 
                                    it will be added as an empty bar categories can be removed from the visuals 

    Returns:
        fig: Matplotlib Figure
        ax: Matplotlib Axes
    
  Source: https://login.codingdojo.com/m/606/13765/117605"""
  # Making our figure with gridspec for subplots
  gridspec = {'height_ratios':[0.7,0.3]}
  fig, axes = plt.subplots(nrows=2, figsize=figsize,
                           sharex=True, gridspec_kw=gridspec)
  # Histogram on Top
  sns.histplot(data=df, x=x, ax=axes[0])
  # Boxplot on Bottom
  sns.boxplot(data=df, x=x, ax=axes[1])
  ## Adding a title
  axes[0].set_title(f"Column: {x}")#, fontweight='bold')
  ## Adjusting subplots to best fill Figure
  fig.tight_layout()

  # Ensure plot is shown before message
  if show:
      plt.show()
   
  ## Print message with info on the count and % of null values
  null_count = df[x].isna().sum()
  null_perc = null_count/len(df)* 100
  print(f"- NaN's Found: {null_count} ({round(null_perc,2)}%)")
  return fig, axes




def explore_categorical(df, x, fillna = True, placeholder = 'MISSING',
                        figsize = (6,4), order = None, show=True):
    """Plots a seaborn countplot of for x column and prints information on:
    - the # and % of null values
    - number of unique values
    - the most frequent category  and how much of the feature is this category (%)
    - A warning message if the feature is quasi-constant or constant feature
                            (if more than 99% of feature is a single value)

    Args:
        df (Frame): DataFrame that contains column x
        x (str): a column name 
        fillna (bool, optional): if True, fillna with the placeholder. Defaults to True.
        placeholder (str, optional): Value used to fillna if fillna is True. Defaults to 'MISSING'.
        figsize (tuple, optional): Figure size (width, height). Defaults to (6,4).
        order (list, optional): List of categories to include in graph, in the specified order. Defaults to None. 
                                Note: any category not in the order list will not be shown on the graph.
                                    If a category is included in the order list that isn't in the data, 
                                    it will be added as an empty bar categories can be removed from the visuals 

    Returns:
        fig: Matplotlib Figure
        ax: Matplotlib Axes
    """
    # Make a copy of the dataframe and fillna 
    temp_df = df.copy()


    ## Save null value counts and percent for printing 
    null_count = temp_df[x].isna().sum()
    null_perc = null_count/len(temp_df)* 100


    # fillna with placeholder
    if fillna == True:
        temp_df[x] = temp_df[x].fillna(placeholder)


    # Create figure with desired figsize
    fig, ax = plt.subplots(figsize=figsize)

    ## Plotting a count plot 
    sns.countplot(data=temp_df, x=x, ax=ax, order=order)

    # Rotate Tick Labels for long names
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha='right')

    # Add. title with the feature name included
    ax.set_title(f"Column: {x}")#, fontweight='bold')

    # Fix layout and show plot (before print statements)
    fig.tight_layout()
    if show:
        plt.show()


    # Print null value info
    print(f"- NaN's Found: {null_count} ({round(null_perc,2)}%)")
    # Print cardinality info
    nunique = temp_df[x].nunique()
    print(f"- Unique Values: {nunique}")


    # Get the most most common value, its count as # and as %
    most_common_val_count = temp_df[x].value_counts(dropna=False).head(1)
    most_common_val = most_common_val_count.index[0]
    freq = most_common_val_count.values[0]

    perc_most_common = freq / len(temp_df) * 100
    print(f"- Most common value: '{most_common_val}' occurs {freq} times ({round(perc_most_common,2)}%)")

    # print message if quasi-constant or constant (most common val more than 98% of data)
    if perc_most_common > 98:
        print(f"\n- [!] Warning: '{x}' is a constant or quasi-constant feature and should be dropped.")

    return fig, ax



def plot_categorical_vs_target(df, x, y,
                                   target_type='reg',
                                   figsize=(6,4),
                                   fillna = True, placeholder = 'MISSING',
                                   order = None, show=True
                                   ):
  """Updated Version of the function which accepts either numeric or categorical targets.
  Adapted from Source: https://login.codingdojo.com/m/606/13765/117606
  Plots a combination seaborn barplot (without error bars) and a stripplot.

    Args:
        df (Frame): DataFrame containing data to plot.
        x (str): Column to use as the x-axis (categories)
        y (str, optional): Target column to plot on the y-axis. Defaults to 'SalePrice'.		
        fillna (bool, optional): if True, fillna with the placeholder. Defaults to True.
        placeholder (str, optional): Value used to fillna if fillna is True. Defaults to 'MISSING'.
        figsize (tuple, optional): Figure size (width, height). Defaults to (6,4).
        order (list, optional): List of categories to include in graph, in the specified order. Defaults to None. 
                                Note: any category not in the order list will not be shown on the graph.
                                    If a category is included in the order list that isn't in the data, 
                                    it will be added as an empty bar categories can be removed from the visuals 

    Returns:
        fig: Matplotlib Figure
        ax: Matplotlib Axes
    """
 
  # Make a copy of the dataframe and fillna
  temp_df = df.copy()
  # fillna with placeholder
  if fillna == True:
    temp_df[x] = temp_df[x].fillna(placeholder)

  # or drop nulls prevent unwanted 'nan' group in stripplot
  else:
    temp_df = temp_df.dropna(subset=[x])


  # Create the figure and subplots
  fig, ax = plt.subplots(figsize=figsize)

  ## If a regression target:
  if 'reg' in target_type:

      # Barplot
    sns.barplot(data=temp_df, x=x, y=y, ax=ax, order=order, alpha=0.6,
                linewidth=1, edgecolor='black', errorbar=None)

    # Boxplot
    sns.stripplot(data=temp_df, x=x, y=y, hue=x, ax=ax,
                  order=order, hue_order=order, legend=False,
                  edgecolor='white', linewidth=0.5,
                  size=3,zorder=0)
    # Rotate xlabels
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha='right')


  # If a classification target:
  elif 'class' in target_type:
   sns.histplot(data=df, hue=y, x=x,hue_order=order,
                  stat='percent', multiple='fill',ax=ax)
  else:
    raise Exception("target_type must be one either 'class' or 'reg'")

  # Final Plot customization
  # Add a title
  ax.set_title(f"{x} vs. {y}")#, fontweight='semibold')
  fig.tight_layout()
  if show==True:
      plt.show()
  return fig, ax



def plot_numeric_vs_target(df, x, y, figsize=(6,4),
                           target_type='reg', errorbar='ci',
                           estimator='mean', order=None,show=True,
                           **kwargs): # kwargs for sns.regplot
  """UPDATED FUNCTION WITH OPTION FOR WHICH TYPE OF TARGET
  Source: https://login.codingdojo.com/m/606/13765/117605
  Plots a seaborn regplot, with an optional formula annotation.
    Also calculates correlation and displays Pearson's r in the title.

    Args:
        df (Frame): DataFrame with data.
        x (str): Numeric column name.
        y (str, optional): Numeric target column name. Defaults to 'SalePrice'.
        figsize (tuple, optional): Figure size. Defaults to (6,4).
        annotate (bool, optional): Whether to annotate regplot equation. Defaults to False. 
        
    Returns:
        fig: Matplotlib Figure
        ax: Matplotlib Axes
  """

  nulls = df[[x,y]].isna().sum()
  if nulls.sum()>0:
    print(f"- Excluding {nulls.sum()} NaN's")
    # print(nulls)
    temp_df = df.dropna(subset=[x,y,])
  else:
    temp_df = df

  if 'reg' in target_type:
    fig, axes = plt.subplots(figsize=figsize)
    # Calculate the correlation
    corr = df[[x,y]].corr().round(2)
    r = corr.loc[x,y]
    # Plot the data
    scatter_kws={'ec':'white','lw':1,'alpha':0.8}
    sns.regplot(data=temp_df, x=x, y=y, ax=axes, scatter_kws=scatter_kws, **kwargs) # Included the new argument within the sns.regplot function
    ## Add the title with the correlation
    axes.set_title(f"{x} vs. {y} (r = {r})")#, fontweight='bold')


  elif 'class' in target_type:
    fig, axes = plt.subplots(figsize=figsize, ncols=2)

    # Left Subplot (barplot)
    sns.barplot(data=temp_df, x=y, y=x,  order=order, estimator=estimator,
                errorbar=errorbar, ax=axes[0],)

    ## Right subplot (boxplot+stripplot)
    # Stripplot
    sns.stripplot(data=temp_df, x=y, y=x, hue=y, ax=axes[1],
                order=order, hue_order=order, legend=False,
                edgecolor='white', linewidth=0.5,
                size=3,zorder=0)

    # Boxplot
    transparent = {'alpha':.6} #Props for boxplot
    sns.boxplot(data=temp_df, x=y, y=x,
                boxprops=transparent, whiskerprops=transparent,
                width=.25, showfliers=False,
                saturation=0.5,
                ax=axes[1])
    # Add title
    fig.suptitle(f"{x} vs. {y}")

  # Make sure the plot is shown before the print statement
  fig.tight_layout()
  if show==True:
      plt.show()
  return fig, axes



######### NEW
def plot_correlation(df, cmap='coolwarm', cols=None):
  """
  Plots a correlation matrix heatmap for the given DataFrame.

  Parameters:
  df (DataFrame): The input DataFrame.
  cmap (str, optional): The color map to use for the heatmap. Defaults to 'coolwarm'.
  cols (list, optional): The columns to include in the correlation matrix. If None, all columns are included.

  Returns:
  fig (Figure): The matplotlib Figure object containing the correlation matrix heatmap.
  """
  if cols == None:
    cols = df.columns
  corr = df[cols].corr(numeric_only=True)

  fig, ax = plt.subplots(figsize=(8,6))
  sns.heatmap(corr, cmap=cmap, ax=ax, annot=True, center=0)
  ax.set_title("Correlation Matrix")
  return fig


from ._eda_functions_plotly import *



def annotate_regplot_equation(ax):
  """
  Annotates a regression plot with the equation of the regression line.

  Parameters:
  ax (matplotlib.axes.Axes): The axes object containing the regression plot.

  Adapted from Source: https://www.statology.org/seaborn-regplot-equation/

  Example Use:
  >> fig, ax = plot_numeric_vs_target(df, x="Living Area Sqft")
  >> annotate_regplot_equation(ax)
  """
  import scipy
  #calculate slope and intercept of regression equation
  slope, intercept, r, p, sterr = scipy.stats.linregress(x=ax.get_lines()[0].get_xdata(),
                            y=ax.get_lines()[0].get_ydata())
  eqn = f'y = {slope:,.2f} * X + {intercept:,.2f}'
  ax.legend(handles=[ax.get_lines()[0]], labels=[eqn])


"""
nlp functions from Coding Dojo Online Data Science and Machine Learning 24-Week Program.
"""


def reference_set_seed_keras(markdown=True):
	
	ref = """
	```python
	# From source: https://keras.io/examples/keras_recipes/reproducibility_recipes/
	import tensorflow as tf
	import numpy as np

	# Then Set Random Seeds
	tf.keras.utils.set_random_seed(42)
	tf.random.set_seed(42)
	np.random.seed(42)

	# Then run the Enable Deterministic Operations Function
	tf.config.experimental.enable_op_determinism()
	```
	"""
	if markdown:
		from IPython.display import display, Markdown
		display(Markdown(ref))
	else:
		print(ref)


def make_text_vectorization_layer(train_ds,  max_tokens=None, 
								  split='whitespace',
								  standardize="lower_and_strip_punctuation",
								  output_mode="int",
								  output_sequence_length=None,
								  ngrams=None, pad_to_max_tokens=False,
								  verbose=True,
								  **kwargs,
								 ):
	"""
	Creates a text vectorization layer using TensorFlow's TextVectorization class.

	Parameters:
	- train_ds: The training dataset containing the text data.
	- max_tokens: The maximum number of tokens to keep in the vocabulary.
	- split: The method used to split the text into tokens.
	- standardize: The method used to standardize the text.
	- output_mode: The output mode of the layer.
	- output_sequence_length: The length of the output sequences.
	- ngrams: The n-grams to consider when tokenizing the text.
	- pad_to_max_tokens: Whether to pad the sequences to have the same length.
	- verbose: Whether to print the layer's parameters.
	- **kwargs: Additional keyword arguments to pass to the TextVectorization class.

	Returns:
	- text_vectorizer: The text vectorization layer.
	- int_to_str: A dictionary mapping integers to words in the vocabulary.
	"""
	import tensorflow as tf
	import numpy as np
	from pprint import pprint

	# Build the text vectorization layer
	text_vectorizer = tf.keras.layers.TextVectorization(
		max_tokens=max_tokens,
		standardize=standardize, 
		output_mode=output_mode,
		output_sequence_length=output_sequence_length,
		**kwargs
	)
	# Get just the text from the training data
	
	if isinstance(train_ds, (np.ndarray, list, tuple, pd.Series)):
		ds_texts = train_ds
	else:
		try:
			ds_texts = train_ds.map(lambda x, y: x )
		except:
			ds_texts = train_ds
			
	# Fit the layer on the training texts
	text_vectorizer.adapt(ds_texts)
	
	
	if verbose:
		# Print the params
		print( "\ntf.keras.layers.TextVectorization(" )
		config = text_vectorizer.get_config()
		pprint(config,indent=4)
		print(")")
			   
	# SAVING VOCAB FOR LATER
	# Getting list of vocab 
	vocab = text_vectorizer.get_vocabulary()
	# Save dictionaries to look up words from ints 
	int_to_str  = {idx:word for idx, word in enumerate(vocab)}
	
	return text_vectorizer, int_to_str


def batch_preprocess_texts(
	texts,
	nlp=None,
	remove_stopwords=True,
	remove_punct=True,
	use_lemmas=False,
	disable=["ner"],
	batch_size=50,
	n_process=-1,
):
	"""Efficiently preprocess a collection of texts using nlp.pipe()

	Args:
		texts (collection of strings): Collection of texts to process (e.g. df['text'])
		nlp (spacy pipe), optional): Spacy nlp pipe. Defaults to None; if None, it creates a default 'en_core_web_sm' pipe.
		remove_stopwords (bool, optional): Controls stopword removal. Defaults to True.
		remove_punct (bool, optional): Controls punctuation removal. Defaults to True.
		use_lemmas (bool, optional): Lemmatize tokens. Defaults to False.
		disable (list of strings, optional): Named pipeline elements to disable. Defaults to ["ner"]: Used with nlp.pipe(disable=disable)
		batch_size (int, optional): Number of texts to process in a batch. Defaults to 50.
		n_process (int, optional): Number of CPU processors to use. Defaults to -1 (meaning all CPU cores).

	Returns:
		list of tokens: Processed texts as a list of tokens.
	"""
	# Function implementation


import pandas as pd
def get_ngram_measures_finder(tokens, ngrams=2, measure='raw_freq', top_n=None, min_freq=1, words_colname='Words'):
	"""
	Calculate n-gram measures for a given list of tokens.

	Parameters:
	- tokens (list): List of tokens.
	- ngrams (int): Number of grams to consider (2 for bigrams, 3 for trigrams, 4 for quadgrams). Default is 2.
	- measure (str): Measure to calculate ('raw_freq' for raw frequency, 'pmi' for pointwise mutual information). Default is 'raw_freq'.
	- top_n (int): Number of top n-grams to return. Default is None (return all n-grams).
	- min_freq (int): Minimum frequency threshold for n-grams. Default is 1.
	- words_colname (str): Column name for the n-grams in the resulting DataFrame. Default is 'Words'.

	Returns:
	- df_ngrams (DataFrame): DataFrame containing the n-grams and their corresponding measure values.
	"""
	import nltk
	import pandas as pd

	if ngrams == 4:
		MeasuresClass = nltk.collocations.QuadgramAssocMeasures
		FinderClass = nltk.collocations.QuadgramCollocationFinder
	elif ngrams == 3:
		MeasuresClass = nltk.collocations.TrigramAssocMeasures
		FinderClass = nltk.collocations.TrigramCollocationFinder
	else:
		MeasuresClass = nltk.collocations.BigramAssocMeasures
		FinderClass = nltk.collocations.BigramCollocationFinder

	measures = MeasuresClass()

	finder = FinderClass.from_words(tokens)
	finder.apply_freq_filter(min_freq)
	if measure == 'pmi':
		scored_ngrams = finder.score_ngrams(measures.pmi)
	else:
		measure = 'raw_freq'
		scored_ngrams = finder.score_ngrams(measures.raw_freq)

	df_ngrams = pd.DataFrame(scored_ngrams, columns=[words_colname, measure.replace("_", ' ').title()])
	if top_n is not None:
		return df_ngrams.head(top_n)
	else:
		return df_ngrams

# def get_ngram_measures_finder(tokens=None,docs=None, ngrams=2, verbose=False,
#                               get_scores_df=False, measure='raw_freq', top_n=None,
#                              words_colname='Words'):
#     import nltk
#     if ngrams == 4:
#         MeasuresClass = nltk.collocations.QuadgramAssocMeasures
#         FinderClass = nltk.collocations.QuadgramCollocationFinder
		
#     elif ngrams == 3: 
#         MeasuresClass = nltk.collocations.TrigramAssocMeasures
#         FinderClass = nltk.collocations.TrigramCollocationFinder
#     else:
#         MeasuresClass = nltk.collocations.BigramAssocMeasures
#         FinderClass = nltk.collocations.BigramCollocationFinder

#     measures = MeasuresClass()
	
#     if (tokens is not None):
#         finder = FinderClass.from_words(tokens)
#     elif (docs is not None):
#         finder = FinderClass.from_docs(docs)
#     else:
#         raise Exception("Must provide tokens or docs")
		

#     if get_scores_df == False:
#         return measures, finder
#     else:
#         df_ngrams = get_score_df(measures, finder, measure=measure, top_n=top_n, words_colname=words_colname)
#         return df_ngrams




# def get_score_df( measures,finder, measure='raw_freq', top_n=None, words_colname="Words"):
#     import pandas as pd
#     if measure=='pmi':
#         scored_ngrams = finder.score_ngrams(measures.pmi)
#     else:
#         measure='raw_freq'
#         scored_ngrams = finder.score_ngrams(measures.raw_freq)

#     df_ngrams = pd.DataFrame(scored_ngrams, columns=[words_colname, measure.replace("_",' ').title()])
#     if top_n is not None:
#         return df_ngrams.head(top_n)
#     else:
#         return df_ngrams


	

def preprocess_text(txt, nlp=None, remove_stopwords=True, remove_punct=True, use_lemmas=False,):
	"""
	Preprocesses the given text by tokenizing and optionally removing stopwords, punctuation, and lemmatizing the tokens.

	Args:
		txt (str): The text to be processed.
		nlp (spacy pipe, optional): The Spacy nlp pipe. Defaults to None.
								   If None, it creates a default 'en_core_web_sm' pipe.
		remove_stopwords (bool, optional): Controls whether to remove stopwords. Defaults to True.
		remove_punct (bool, optional): Controls whether to remove punctuation. Defaults to True.
		use_lemmas (bool, optional): Controls whether to lemmatize tokens. Defaults to False.

	Returns:
		list: A list of tokens/lemmas after preprocessing.
	"""
	import spacy
	if nlp is None:
		nlp = spacy.load('en_core_web_sm')

	doc = nlp(txt)

	# Saving list of the token objects for stopwords and punctuation removal
	tokens = []

	for token in doc:
		# Check if should remove stopwords and if token is stopword
		if (remove_stopwords == True) & (token.is_stop == True):
			# Continue the loop with the next token
			continue

		# Check if should remove punctuation and if token is punctuation
		if (remove_punct == True) & (token.is_punct == True):
			# Continue the loop with the next token
			continue

		# Check if should remove punctuation and if token is a space
		if (remove_punct == True) & (token.is_space == True):
			# Continue the loop with the next token
			continue

		# Determine final form of output list of tokens/lemmas
		if use_lemmas:
			tokens.append(token.lemma_.lower())
		else:
			tokens.append(token.text.lower())

	return tokens





def make_custom_nlp(
	disable=["ner"],
	contractions=["don't", "can't", "couldn't", "you'd", "I'll"],
	stopwords_to_add=[],
	stopwords_to_remove=[],
	spacy_model = "en_core_web_sm"
):
	"""Returns a custom spacy nlp pipeline.
	
	Args:
		disable (list, optional): Names of pipe components to disable. Defaults to ["ner"].
		contractions (list, optional): List of contractions to add as special cases. Defaults to ["don't", "can't", "couldn't", "you'd", "I'll"].
		stopwords_to_add(list, optional): List of words to set as stopwords (word.is_stop=True)
		stopwords_to_remove(list, optional): List of words to remove from stopwords (word.is_stop=False)
		spacy_model (str, optional): Name or path of the spaCy model to load. Defaults to "en_core_web_sm".
			
	Returns:
		nlp pipeline: spacy pipeline with special cases and updated nlp.Default.stopwords
	"""
	import spacy
	# Load the English NLP model
	nlp = spacy.load(spacy_model, disable=disable)
	

	## Adding Special Cases 
	# Loop through the contractions list and add special cases
	for contraction in contractions:
		special_case = [{"ORTH": contraction}]
		nlp.tokenizer.add_special_case(contraction, special_case)

	
	## Adding stopwords
	for word in stopwords_to_add:
		# Set the is_stop attribute for the word in the vocab dict to true.
		nlp.vocab[
			word
		].is_stop = True  # this determines spacy's treatmean of the word as a stop word

		# Add the word to the list of stopwords (for easily tracking stopwords)
		nlp.Defaults.stop_words.add(word)

	
	## Removing Stopwords
	for word in stopwords_to_remove:
		
		# Ensure the words are not recognized as stopwords
		nlp.vocab[word].is_stop = False
		nlp.Defaults.stop_words.discard(word)
		
	return nlp




# def custom_preprocess_text(
# 	txt,
# 	nlp=None,
# 	nlp_creation_fn=None,
# 	nlp_fn_kwargs={},  ## THESE ARE NEW SINCE BRENDA SAW
# 	lowercase=True,
# 	remove_stopwords=True,
# 	remove_punct=True,
# 	use_lemmas=False,
# 	disable=None,
# ):
# 	"""Preprocess text into tokens/lemmas. 
	
# 	Args:
# 		txt (string): text to process
# 		nlp (spacy pipe), optional): Spacy nlp pipe. Defaults to None; if None, it creates a default 'en_core_web_sm' pipe.
# 		nlp_creation_fn (_type_, optional): Function that returns an nlp pipe. Defaults to None; Only used if nlp arg is None.
# 		nlp_fn_kwargs (dict, optional): Keyword arguments for nlp_creation_fn. Defaults to {}.
# 		remove_stopwords (bool, optional): Controls stopword removal. Defaults to True.
# 		remove_punct (bool, optional): Controls punctuation removal. Defaults to True.
# 		use_lemmas (bool, optional): lemmatize tokens. Defaults to False.
# 		disable (list of strings, optional): named pipeline elements to disable. Defaults to None;Only used if nlp is None and nlp_creation_fn is None
	
# 	Returns:
# 		list: list of tokens/lemmas
# 	"""
# 	# If nlp is none, use nlp_creation_func to make it
# 	if (nlp is None) and (nlp_creation_fn is not None):
# 		nlp = nlp_creation_fn(**nlp_fn_kwargs)
	
# 	# If nlp is none,and no nlp_creation_func, make default nlp object
# 	elif (nlp is None) & (nlp_creation_fn is None):
# 		nlp = spacy.load("en_core_web_sm")

# 	# Create the document
# 	doc = nlp(txt)
	
# 	# Saving list of the token objects for stopwords and punctuation removal
# 	tokens = []
	
# 	for token in doc:
# 		# Check if should remove stopwords and if token is stopword
# 		if (remove_stopwords == True) & (token.is_stop == True):
# 			# Continue the loop with the next token
# 			continue
	
# 		# Check if should remove punctuation and if token is punctuation
# 		if (remove_punct == True) & (token.is_punct == True):
# 			# Continue the loop with the next oken
# 			continue

# 		# Check if should remove punctuation and if token is a space
# 		if (remove_punct == True) & (token.is_space == True):
# 			# Continue the loop with the next oken
# 			continue
	
# 		## Determine final form of output list of tokens/lemmas
# 		if use_lemmas:
# 			tokens.append(token.lemma_)
# 		elif lowercase==True:
# 			tokens.append(token.text.lower())
# 		else: 
# 			tokens.append(token.text)
	
# 	return tokens





###### PLOTLY VERSIONS

def set_template(template='seaborn'):
    """
    Sets the default template for Plotly plots.

    Parameters:
    - template (str): The name of the template to set. Default is 'seaborn'.
    """
    import plotly.io as pio
    pio.templates.default = template
    

def plotly_explore_numeric(df, x, width=1000, height=500):
    """
    Generate a histogram with a box plot overlay to explore the distribution of a numeric variable.

    Parameters:
    - df: DataFrame - The input DataFrame containing the data.
    - x: str - The name of the column representing the numeric variable to be explored.
    - width: int, optional - The width of the generated plot in pixels. Default is 1000.
    - height: int, optional - The height of the generated plot in pixels. Default is 500.

    Returns:
    - fig: plotly.graph_objects.Figure - The generated plotly figure object.
    """
    import plotly.express as px
    fig = px.histogram(df, x=x, marginal='box', title=f'Distribution of {x}', 
                       width=width, height=height)
    return fig


def plotly_explore_categorical(df, x ,width=1000, height=500):
    """
    Generate a histogram plot using Plotly to explore the distribution of a categorical variable in a DataFrame.

    Parameters:
    - df (pandas.DataFrame): The DataFrame containing the data.
    - x (str): The name of the categorical variable to explore.
    - width (int, optional): The width of the plot in pixels. Default is 1000.
    - height (int, optional): The height of the plot in pixels. Default is 500.

    Returns:
    - fig (plotly.graph_objects.Figure): The generated histogram plot.
    """
    import plotly.express as px
    fig = px.histogram(df,x=x,color=x,title=f'Distribution of {x}', 
                         width=width, height=height)
    return fig
    

def plotly_numeric_vs_target(df, x, y='SalePrice', trendline='ols', add_hoverdata=True,
                            width=800, height=600):
    """
    Creates a scatter plot using Plotly to visualize the relationship between a numeric variable and a target variable.

    Parameters:
    - df (pandas.DataFrame): The dataframe containing the data.
    - x (str): The name of the numeric variable to be plotted on the x-axis.
    - y (str): The name of the target variable to be plotted on the y-axis. Default is 'SalePrice'.
    - trendline (str): The type of trendline to be added to the plot. Default is 'ols' (ordinary least squares).
    - add_hoverdata (bool): Whether to include hover data in the plot. Default is True.
    - width (int): The width of the plot in pixels. Default is 800.
    - height (int): The height of the plot in pixels. Default is 600.

    Returns:
    - pfig (plotly.graph_objects.Figure): The Plotly figure object representing the scatter plot.
    """
    if add_hoverdata == True:
        hover_data = list(df.drop(columns=[x, y]).columns)
    else: 
        hover_data = None
    import plotly.express as px
    pfig = px.scatter(df, x=x, y=y, hover_data=hover_data,
                      trendline=trendline, trendline_color_override='red',
                      title=f"{x} vs. {y}", width=width, height=height)

    pfig.update_traces(marker=dict(size=3), line=dict(dash='dash'))
    return pfig


def plotly_numeric_vs_target(df, x, y='SalePrice', trendline='ols', add_hoverdata=True,
                            width=800, height=600):
    """
    Creates a scatter plot using Plotly to visualize the relationship between a numeric variable and a target variable.

    Parameters:
    - df (pandas.DataFrame): The dataframe containing the data.
    - x (str): The name of the numeric variable to be plotted on the x-axis.
    - y (str, optional): The name of the target variable to be plotted on the y-axis. Default is 'SalePrice'.
    - trendline (str, optional): The type of trendline to be added to the plot. Default is 'ols' (ordinary least squares).
    - add_hoverdata (bool, optional): Whether to include hover data in the plot. Default is True.
    - width (int, optional): The width of the plot in pixels. Default is 800.
    - height (int, optional): The height of the plot in pixels. Default is 600.

    Returns:
    - pfig (plotly.graph_objects.Figure): The Plotly figure object representing the scatter plot.
    """
    if add_hoverdata == True:
        hover_data = list(df.drop(columns=[x, y]).columns)
    else: 
        hover_data = None
    import plotly.express as px
    pfig = px.scatter(df, x=x, y=y, hover_data=hover_data,
                      trendline=trendline, trendline_color_override='red',
                      title=f"{x} vs. {y}", width=width, height=height)

    pfig.update_traces(marker=dict(size=3), line=dict(dash='dash'))
    return pfig
def plotly_numeric_vs_target(df, x, y='SalePrice', trendline='ols',add_hoverdata=True,
                            width=800, height=600,):
    if add_hoverdata == True:
        hover_data = list(df.drop(columns=[x,y]).columns)
    else: 
        hover_data = None
    import plotly.express as px
    pfig = px.scatter(df, x=x, y=y,# template='plotly_dark', 
                     hover_data=hover_data,
                      trendline=trendline,
                      trendline_color_override='red',
                     title=f"{x} vs. {y}",
                      width=width, height=height)

    
    pfig.update_traces(marker=dict(size=3),
                      line=dict(dash='dash'))
    return pfig
    

# def plotly_categorical_vs_target(df, x, y='SalePrice', agg='mean',width=800,height=500,):
#     if agg=='mean':
#         plot_df = df.groupby(x,as_index=False)[y].mean().sort_values(y, ascending=False)
        
#     elif agg=='median':
#         plot_df = df.groupby(x,as_index=False)[y].median().sort_values(y,ascending=False)
        
#     else:
#         plot_df = df
        
#     fig = px.bar(plot_df, x=x,y=y, color=x, title=f'Compare {agg.title()} {y} by {x}',
#                  width=width, height=height)

                
#     return fig

def plotly_categorical_vs_target(df, x, y='SalePrice', histfunc='avg', width=800,height=500):
    """
    Plots a categorical variable against a target variable using Plotly.

    Parameters:
    - df (pandas.DataFrame): The DataFrame containing the data.
    - x (str): The name of the categorical variable.
    - y (str): The name of the target variable. Default is 'SalePrice'.
    - histfunc (str): The aggregation function to use for the histogram. Default is 'avg'.
    - width (int): The width of the plot. Default is 800.
    - height (int): The height of the plot. Default is 500.

    Returns:
    - fig (plotly.graph_objects.Figure): The Plotly figure object.
    """
    import plotly.express as px
    fig = px.histogram(df, x=x,y=y, color=x, width=width, height=height,
                       histfunc=histfunc, title=f'Compare {histfunc.title()} {y} by {x}')
    fig.update_layout(showlegend=False)
    return fig


def plotly_correlation(df, cmap='magma', cols=None):
    """
    Generates a correlation heatmap using Plotly.

    Parameters:
    df (pandas.DataFrame): The input DataFrame.
    cmap (str, optional): The color map to use for the heatmap. Defaults to 'magma'.
    cols (list, optional): The columns to include in the correlation calculation. If None, all columns are included. Defaults to None.

    Returns:
    plotly.graph_objects.Figure: The correlation heatmap figure.
    """
    import plotly.express as px

    if cols == None:
        cols = df.columns
    corr = df[cols].corr(numeric_only=True).round(2)
    fig = px.imshow(corr, text_auto=True,width=600, height=600,
                    color_continuous_scale=cmap,
                    color_continuous_midpoint=0, 
                   title='Correlation Heatmap')
    return fig


# if __name__ == "__main__":
#     set_template()


## PREVIOUS CLASSIFICATION_METRICS FUNCTION FROM INTRO TO ML
def classification_metrics(y_true, y_pred, label='',
                           output_dict=False, figsize=(8,4),
                           normalize='true', cmap='Blues',
                           colorbar=False, values_format=".2f",
                           target_names = None, return_fig=True):
    """
    Calculate classification metrics from predictions and display Confusion matrix.

    Args:
        y_true (Series/array): True target values.
        y_pred (Series/array): Predicted target values.
        label (str, optional): Label for printed header. Defaults to ''.
        output_dict (bool, optional): Return the results of classification_report as a dict. Defaults to False.
        figsize (tuple, optional): figsize for confusion matrix subplots. Defaults to (8,4).
        normalize (str, optional): Arg for sklearn's ConfusionMatrixDisplay. Defaults to 'true' (conf mat values normalized to true class).
        cmap (str, optional): Colormap for the ConfusionMatrixDisplay. Defaults to 'Blues'.
        colorbar (bool, optional): Arg for ConfusionMatrixDisplay: include colorbar or not. Defaults to False.
        values_format (str, optional): Format values on confusion matrix. Defaults to ".2f".
        target_names (array, optional): Text labels for the integer-encoded target. Passed in numeric order [label for "0", label for "1", etc.].
        return_fig (bool, optional): To get matplotlib figure for confusion matrix, set output_dict to False and set return_fig to True.

    Returns:
        dict: Dictionary from classification_report. Only returned if output_dict=True.
        fig: Matplotlib figure with confusion matrix. Only returned if output_dict=False and return_fig=True.

    Note: 
        This is a modified version of the classification metrics function from Intro to Machine Learning.
        Updates:
          - Reversed raw counts confusion matrix cmap (so darker==more).
          - Added arg for normalized confusion matrix values_format.
    """
    from sklearn.metrics import classification_report, ConfusionMatrixDisplay
    import matplotlib.pyplot as plt
    import numpy as np
    # Get the classification report
    report = classification_report(y_true, y_pred,target_names=target_names)
    
    ## Print header and report
    header = "-"*70
    print(header, f" Classification Metrics: {label}", header, sep='\n')
    print(report)
    
    ## CONFUSION MATRICES SUBPLOTS
    fig, axes = plt.subplots(ncols=2, figsize=figsize)
    
    # Create a confusion matrix of raw counts (left subplot)
    ConfusionMatrixDisplay.from_predictions(y_true, y_pred,
                                            normalize=None, 
                                            cmap='gist_gray_r',# Updated cmap
                                            values_format="d", 
                                            colorbar=colorbar,
                                            ax = axes[0], 
                                            display_labels=target_names);
    axes[0].set_title("Raw Counts")
    
    # Create a confusion matrix with the data with normalize argument 
    ConfusionMatrixDisplay.from_predictions(y_true, y_pred,
                                            normalize=normalize,
                                            cmap=cmap, 
                                            values_format=values_format, #New arg
                                            colorbar=colorbar,
                                            ax = axes[1],
                                            display_labels=target_names);
    axes[1].set_title("Normalized Confusion Matrix")
    
    # Adjust layout and show figure
    fig.tight_layout()
    plt.show()
    
    # Return dictionary of classification_report
    if output_dict==True:
        report_dict = classification_report(y_true, y_pred,target_names=target_names, output_dict=True)
        return report_dict

    elif return_fig == True:
        return fig
    
    
    
def evaluate_classification(model, X_train=None, y_train=None, X_test=None, y_test=None,
                            figsize=(6,4), normalize='true', output_dict = False,
                            cmap_train='Blues', cmap_test="Reds",colorbar=False,
                            values_format='.2f',
                            target_names=None, return_fig=False):
    """Evalutes an sklearn-compatible classification model on training and test data. 
    For each data split, return the classification report and confusion matrix display. 

    Args:
        model (sklearn estimator): Classification model to evaluate.
        X_train (Frame/Array, optional): Training data. Defaults to None.
        y_train (Series/Array, optional): Training labels. Defaults to None.
        X_test (Frame/Array, optional): Test data. Defaults to None.
        y_test (Series/Array, optional): Test labels. Defaults to None.
        figsize (tuple, optional): figsize for confusion matrix subplots. Defaults to (6,4).
        normalize (str, optional): arg for sklearn's ConfusionMatrixDisplay. Defaults to 'true' (conf mat values normalized to true class).  
        output_dict (bool, optional):  Return the results of classification_report as a dict. Defaults to False. Defaults to False.
        cmap_train (str, optional): Colormap for the ConfusionMatrixDisplay for training data. Defaults to 'Blues'.
        cmap_test (str, optional): Colormap for the ConfusionMatrixDisplay for test data.  Defaults to "Reds".
        colorbar (bool, optional): Arg for ConfusionMatrixDispaly: include colorbar or not. Defaults to False.
        values_format (str, optional): Format values on confusion matrix. Defaults to ".2f".
        target_names (array, optional): Text labels for the integer-encoded target. Passed in numeric order [label for "0", label for "1", etc.]
        return_fig (bool, optional): Whether the matplotlib figure for confusion matrix is returned. Defaults to False.
                                          Note: Must set outout_dict to False and set return_fig to True to get figure returned.
     Returns (Only 1 value is returned, but contents vary):
        dict: Dictionary that contains results for "train" and "test. 
              Contents of dictionary depending on output_dict and return_fig:
              - if output_dict==True and return_fig==False: returns dictionary of classification report results
            - if output_dict==False and return_fig==True: returns dictionary of confusion matrix displays.
    """
    # Combining arguments used for both training and test results
    shared_kwargs = dict(output_dict=output_dict,  # output_dict: Changed from hard-coded True
                    #   figsize=figsize, 
                      colorbar=colorbar, 
                      target_names=target_names,
                      values_format=values_format,
                      return_fig=return_fig)
 
    if (X_train is None) & (X_test is None):
        raise Exception('\nEither X_train & y_train or X_test & y_test must be provided.')
 
    if (X_train is not None) & (y_train is not None):
        # Get predictions for training data
        y_train_pred = model.predict(X_train)
        # Call the helper function to obtain regression metrics for training data
        results_train = classification_metrics(y_train, y_train_pred, cmap=cmap_train, figsize=figsize,label='Training Data', **shared_kwargs)
        print()
    else:
        results_train=None
  
    if (X_test is not None) & (y_test is not None):
        # Get predictions for test data
        y_test_pred = model.predict(X_test)
        # Call the helper function to obtain regression metrics for test data
        results_test = classification_metrics(y_test, y_test_pred,  cmap=cmap_test, figsize=figsize, label='Test Data' , **shared_kwargs)
    else:
        results_test = None
  
  
    if (output_dict == True) | (return_fig==True):
        # Store results in a dataframe if ouput_frame is True
        results_dict = {'train':results_train,
                        'test': results_test}
        return results_dict


def evaluate_classification_network(model, 
                                    X_train=None, y_train=None, 
                                    X_test=None, y_test=None,
                                    history=None, history_figsize=(6,6),
                                    figsize=(6,4), normalize='true',
                                    output_dict = False,
                                    cmap_train='Blues',
                                    cmap_test="Reds",
                                    values_format=".2f", 
                                    colorbar=False, target_names=None, 
         return_fig=False):
    """Evaluates a neural network classification task using either
    separate X and y arrays or a tensorflow Dataset

    Args:
        model (sklearn-compatible classifier): Model to evaluate.
        X_train (array or tf.data.Dataset, optional): Training data. Defaults to None.
        y_train (array, or None if X_train is a tf Dataset, optional): Training labels (if not using a tf dataset). Defaults to None.
        X_test (array or tf.data.Dataset, optional): Test data. Defaults to None.
        y_test (array, or None if X_test is a tf Dataset, optional): Test labels (if not using a tf Dataset). Defaults to None.
        history (tensorflow history object, optional): History object from model training. Defaults to None.
        history_figsize (tuple, optional): Total figure size for plot_history. Defaults to (6,8).
        figsize (tuple, optional): figsize for confusion matrix subplots. Defaults to (6,4).
        normalize (str, optional): arg for sklearn's ConfusionMatrixDisplay. Defaults to 'true' (conf mat values normalized to true class).  
        output_dict (bool, optional):  Return the results of classification_report as a dict. Defaults to False. Defaults to False.
        cmap_train (str, optional): Colormap for the ConfusionMatrixDisplay for training data. Defaults to 'Blues'.
        cmap_test (str, optional): Colormap for the ConfusionMatrixDisplay for test data.  Defaults to "Reds".
        colorbar (bool, optional): Arg for ConfusionMatrixDispaly: include colorbar or not. Defaults to False.
        values_format (str, optional): Format values on confusion matrix. Defaults to ".2f".
        target_names (array, optional): Text labels for the integer-encoded target. Passed in numeric order [label for "0", label for "1", etc.]
        return_fig (bool, optional): Whether the matplotlib figure for confusion matrix is returned. Defaults to False.
                                          Note: Must set outout_dict to False and set return_fig to True to get figure returned.


     Returns (Only 1 value is returned, but contents vary):
        dict: Dictionary that contains results for "train" and "test. 
            Contents of dictionary depending on output_dict and return_fig:
              - if output_dict==True and return_fig==False: returns dictionary of classification report results
            - if output_dict==False and return_fig==True: returns dictionary of confusion matrix displays.
    """
    if (X_train is None) & (X_test is None):
        raise Exception('\nEither X_train & y_train or X_test & y_test must be provided.')
 
    shared_kwargs = dict(output_dict=True, 
                      figsize=figsize,
                      colorbar=colorbar,
                      values_format=values_format, 
                      target_names=target_names,)
    # Plot history, if provided
    if history is not None:
        plot_history(history, figsize=history_figsize)
    ## Adding a Print Header
    print("\n"+'='*80)
    print('- Evaluating Network...')
    print('='*80)
    ## TRAINING DATA EVALUATION
    # check if X_train was provided
    if X_train is not None:
        ## Check if X_train is a dataset
        if hasattr(X_train,'map'):
            # If it IS a Datset:
            # extract y_train and y_train_pred with helper function
            y_train, y_train_pred = get_true_pred_labels(model, X_train)
        else:
            # Get predictions for training data
            y_train_pred = model.predict(X_train)
        ## Pass both y-vars through helper compatibility function
        y_train = convert_y_to_sklearn_classes(y_train)
        y_train_pred = convert_y_to_sklearn_classes(y_train_pred)
        
        # Call the helper function to obtain regression metrics for training data
        results_train = classification_metrics(y_train, y_train_pred, cmap=cmap_train,label='Training Data', **shared_kwargs)
        
        ## Run model.evaluate         
        print("\n- Evaluating Training Data:")
        print(model.evaluate(X_train, return_dict=True))
    
    # If no X_train, then save empty list for results_train
    else:
        results_train = None
  
  
    ## TEST DATA EVALUATION
    # check if X_test was provided
    if X_test is not None:
        ## Check if X_train is a dataset
        if hasattr(X_test,'map'):
            # If it IS a Datset:
            # extract y_train and y_train_pred with helper function
            y_test, y_test_pred = get_true_pred_labels(model, X_test)
        else:
            # Get predictions for training data
            y_test_pred = model.predict(X_test)
        ## Pass both y-vars through helper compatibility function
        y_test = convert_y_to_sklearn_classes(y_test)
        y_test_pred = convert_y_to_sklearn_classes(y_test_pred)
        
        # Call the helper function to obtain regression metrics for training data
        results_test = classification_metrics(y_test, y_test_pred, cmap=cmap_test,label='Test Data', **shared_kwargs)
        
        ## Run model.evaluate         
        print("\n- Evaluating Test Data:")
        print(model.evaluate(X_test, return_dict=True))
      
    # If no X_test, then save empty list for results_test
    else:
        results_test = None
      
    if (output_dict == True) | (return_fig==True):
        # Store results in a dataframe if ouput_frame is True
        results_dict = {'train':results_train,
                        'test': results_test}
        return results_dict




def plot_history(history, figsize=(6,8), return_fig=False):
    """Plots the training and validation curves for all metrics in a Tensorflow History object.

    Args:
        history (Tensorflow History): History output from training a neural network.
        figsize (tuple, optional): Total figure size. Defaults to (6,8).
        return_fig (boolean, optional): If true, return figure instead of displaying it with plt.show()

    Returns:
        None or matplotlib.figure.Figure: If return_fig is True, returns the figure object. Otherwise, displays the figure using plt.show().
    """
    import matplotlib.pyplot as plt
    import numpy as np
    # Get a unique list of metrics 
    all_metrics = np.unique([k.replace('val_','') for k in history.history.keys()])
    # Plot each metric
    n_plots = len(all_metrics)
    fig, axes = plt.subplots(nrows=n_plots, figsize=figsize)
    axes = axes.flatten()
    # Loop through metric names add get an index for the axes
    for i, metric in enumerate(all_metrics):
        # Get the epochs and metric values
        epochs = history.epoch
        score = history.history[metric]
        # Plot the training results
        axes[i].plot(epochs, score, label=metric, marker='.')
        # Plot val results (if they exist)
        try:
            val_score = history.history[f"val_{metric}"]
            axes[i].plot(epochs, val_score, label=f"val_{metric}",marker='.')
        except:
            pass
        finally:
            axes[i].legend()
            axes[i].set(title=metric, xlabel="Epoch",ylabel=metric)
   
    # Adjust subplots and show
    fig.tight_layout()
 
    if return_fig:
        return fig
    else:
        plt.show()


def convert_y_to_sklearn_classes(y, verbose=False):
    """
    Helper function to convert neural network outputs to class labels.

    Args:
        y (array/Series): Predictions to convert to classes.
        verbose (bool, optional): Print which preprocessing approach is used. Defaults to False.

    Returns:
        array: Target as 1D class labels
    """
    import numpy as np

    # If already one-dimension
    if np.ndim(y) == 1:
        if verbose:
            print("- y is 1D, using it as-is.")
        return y

    # If 2 dimensions with more than 1 column:
    elif y.shape[1] > 1:
        if verbose:
            print("- y is 2D with >1 column. Using argmax for metrics.")
        return np.argmax(y, axis=1)

    else:
        if verbose:
            print("y is 2D with 1 column. Using round for metrics.")
        return np.round(y).flatten().astype(int)


def get_true_pred_labels(model, ds):
    """Gets the true labels and predicted probabilities from a Tensorflow model and Dataset object.

    Args:
        model (Tensorflow/Keras model): The model to get predictions from.
        ds (tensorflow.data.Dataset): The dataset to iterate through.

    Returns:
        tuple: A tuple containing the true labels and predicted probabilities.
    """
    import numpy as np

    y_true = []
    y_pred_probs = []

    # Loop through the dataset as a numpy iterator
    for images, labels in ds.as_numpy_iterator():

        # Get prediction with batch_size=1
        y_probs = model.predict(images, batch_size=1, verbose=0)
        # Combine previous labels/preds with new labels/preds
        y_true.extend(labels)
        y_pred_probs.extend(y_probs)
    ## Convert the lists to arrays
    y_true = np.array(y_true)
    y_pred_probs = np.array(y_pred_probs)

    return y_true, y_pred_probs



# #### Regression
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score, mean_absolute_percentage_error
def regression_metrics(y_true, y_pred, label='', verbose=True, output_dict=False):
    """
    Calculate MEA, MSE, RMSE, R-Squared and MAPE using the true and predicted labels.

    Args:
        y_true (Series/array): True target values.
        y_pred (Series/array): Predicted target values.
        label (str, optional): Label to display in results header. Defaults to ''.
        verbose (bool, optional): Controls printing of results. Defaults to True.
        output_dict (bool, optional): Return results in a dictionary. Defaults to False.
            (Note one of either verbose or output_dict should be set to True)

    Returns:
        dict: Dictionary of results with keys: 'Label', 'MAE', 'MSE', 'RMSE', 'MAPE', 'R^2'.
            Only returned if output_dict is True.
    """
    import numpy as np
    # Get metrics
    mae = mean_absolute_error(y_true, y_pred)
    mse = mean_squared_error(y_true, y_pred)
    rmse = mean_squared_error(y_true, y_pred, squared=False)
    r_squared = r2_score(y_true, y_pred)
    mape = mean_absolute_percentage_error(y_true, y_pred)
    if (verbose == False) & (output_dict == False):
        raise Exception("At least one of the following arguments must be set to True: output_dict, verbose.")

    if verbose == True:
        # Print Result with Label and Header
        header = "-" * 60
        print(header, f"Regression Metrics: {label}", header, sep='\n')
        print("Relative Comparison Metrics:")
        print(f"- MAE = {mae:,.3f}")
        print(f"- MSE = {mse:,.3f}")
        print(f"- RMSE = {rmse:,.3f}")
        # print('\n')
        print("\nAbsolute Metrics")
        print(f"- MAPE = {mape:,.3f}")
        print(f"- R^2 = {r_squared:,.3f}")
    if output_dict == True:
        metrics = {'Label':label, 'MAE':mae,
                    'MSE':mse, 'RMSE':rmse, 'MAPE':mape,
                    'R^2':r_squared}
        return metrics



def evaluate_regression(reg, X_train, y_train, X_test, y_test, verbose=True, output_frame=False):
    """Evalutes an sklearn-compatible regression model on training and test data. 
    For each data split, return the classification report and confusion matrix display. 

    Args:
        reg (sklearn estimator): Regression model to evaluate.
        X_train (Frame/Array, optional): Training data. Defaults to None.
        y_train (Series/Array, optional): Training labels. Defaults to None.
        X_test (Frame/Array, optional): Test data. Defaults to None.
        y_test (Series/Array, optional): Test labels. Defaults to None.
        verbose (bool, optional): Controls printing of results. Defaults to True. 
        output_dict (bool, optional): Return results in a dictionary. Defaults to False. (Note one of either verbose or output_dict should be set to True)

    Returns:
        dict: Dictionary of results with keys: 'Label','MAE','MSE', 'RMSE', 'MAPE','R^2'. Only returned if output_dict==True.
    """
    # Get predictions for training data
    y_train_pred = reg.predict(X_train)

    # Call the helper function to obtain regression metrics for training data
    results_train = regression_metrics(y_train, y_train_pred, verbose=verbose, output_dict=output_frame, label='Training Data')
    print()

    # Get predictions for test data
    y_test_pred = reg.predict(X_test)

    # Call the helper function to obtain regression metrics for test data
    results_test = regression_metrics(y_test, y_test_pred, verbose=verbose, output_dict=output_frame, label='Test Data')

    # Store results in a dataframe if output_frame is True
    if output_frame:
        import pandas as pd
        results_df = pd.DataFrame([results_train, results_test])
        # Set the label as the index
        results_df = results_df.set_index('Label')
        # Set index.name to none to get a cleaner looking result
        results_df.index.name = None
        # Return the dataframe
        return results_df.round(3)



    
def evaluate_ols(result,X_train_df, y_train, show_summary=True):
    """Plots a Q-Q Plot and residual plot for a statsmodels OLS regression, with option to display summary.

    Args:
        result (statsmodels RegressionResultsWrapper): The result object obtained from fitting the OLS model.
        X_train_df (pandas DataFrame): The training data features.
        y_train (pandas Series): The training data labels.
        show_summary (bool, optional): Whether to display the summary of the regression model. Defaults to True.
    """
    import matplotlib.pyplot as plt
    import statsmodels.api as sm
    try:
        from IPython.display import display
        display(result.summary())
    except:
        pass
    
    ## save residuals from result
    y_pred = result.predict(X_train_df)
    resid = y_train - y_pred
    
    fig, axes = plt.subplots(ncols=2,figsize=(12,5))
    
    ## Normality 
    sm.graphics.qqplot(resid,line='45',fit=True,ax=axes[0]);
    
    ## Homoscedasticity
    ax = axes[1]
    ax.scatter(y_pred, resid, edgecolor='white',lw=1)
    ax.axhline(0,zorder=0)
    ax.set(ylabel='Residuals',xlabel='Predicted Value');
    plt.tight_layout()
    
    
    
def plot_residuals(model,X_test_df, y_test,figsize=(12,5)):
    """Plots a Q-Q Plot and residual plot for a regression model.

    Args:
        model (regression model): The regression model that supports .predict.
        X_test_df (DataFrame): The test data.
        y_test (array-like): The test labels.
        figsize (tuple, optional): The figsize for the regression plots. Defaults to (12,5).
    """

    import matplotlib.pyplot as plt
    import statsmodels.api as sm
    ## Make predictions and calculate residuals
    y_pred = model.predict(X_test_df)
    resid = y_test - y_pred
    
    fig, axes = plt.subplots(ncols=2,figsize=figsize)
    
    ## Normality 
    sm.graphics.qqplot(resid, line='45',fit=True,ax=axes[0]);
    
    ## Homoscedascity
    ax = axes[1]
    ax.scatter(y_pred, resid, edgecolor='white',lw=0.5)
    ax.axhline(0,zorder=0)
    ax.set(ylabel='Residuals',xlabel='Predicted Value');
    plt.tight_layout()
    
    
    
    
    


## FROM MY "From linear to logistic regression" mini-lesson (interview)
from matplotlib import ticker
from sklearn import metrics
import statsmodels.api as sm
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
# from IPython.display import display


def annotate_hbars(ax, ha='left', va='center', size=12, xytext=(4,0),
                  textcoords='offset points'):
    """
    Annotates horizontal bars on a matplotlib Axes object.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to annotate.
    - ha (str): The horizontal alignment of the annotation text. Default is 'left'.
    - va (str): The vertical alignment of the annotation text. Default is 'center'.
    - size (int): The font size of the annotation text. Default is 12.
    - xytext (tuple): The offset of the annotation text from the annotated point. Default is (4, 0).
    - textcoords (str): The coordinate system used for xytext. Default is 'offset points'.
    """
    for bar in ax.patches:
        val = bar.get_width()
        if val < 0:
            x = 0
        else:
            x = val
        bar_ax = bar.get_y() + bar.get_height()/2
        ax.annotate(f"{val:,.2f}", (x, bar_ax), ha=ha, va=va, size=size,
                    xytext=xytext, textcoords=textcoords)




from sklearn import metrics
import pandas as pd

# def evaluate_regression(model, X_train,y_train, X_test, y_test,as_frame=True):
#   """Evaluates a scikit learn regression model using r-squared and RMSE.
#   Returns the results a DataFrame if as_frame is True (Default).
#   """


#   ## Training Data
#   y_pred_train = model.predict(X_train)
#   r2_train = metrics.r2_score(y_train, y_pred_train)
#   rmse_train = metrics.mean_squared_error(y_train, y_pred_train,
#                                           squared=False)
#   mae_train = metrics.mean_absolute_error(y_train, y_pred_train)


#   ## Test Data
#   y_pred_test = model.predict(X_test)
#   r2_test = metrics.r2_score(y_test, y_pred_test)
#   rmse_test = metrics.mean_squared_error(y_test, y_pred_test,
#                                           squared=False)
#   mae_test = metrics.mean_absolute_error(y_test, y_pred_test)

#   if as_frame:
#       df_version =[['Split','R^2','MAE','RMSE']]
#       df_version.append(['Train',r2_train, mae_train, rmse_train])
#       df_version.append(['Test',r2_test, mae_test, rmse_test])
#       df_results = pd.DataFrame(df_version[1:], columns=df_version[0])
#       df_results = df_results.round(2)


#       # adapting hide_index for pd version
#       if pd.__version__ < "1.4.0":
#         display(df_results.style.hide_index().format(precision=2, thousands=','))
#       else:
#         display(df_results.style.hide(axis='index').format(precision=2, thousands=','))

#   else:
#       print(f"Training Data:\tR^2 = {r2_train:,.2f}\tRMSE = {rmse_train:,.2f}\tMAE = {mae_train:,.2f}")
#       print(f"Test Data:\tR^2 = {r2_test:,.2f}\tRMSE = {rmse_test:,.2f}\tMAE = {mae_test:,.2f}")



def get_coefficients(reg, name='Coefficients'):
    """Save a model's .coef_ and .intercept_ as a Pandas Series"""
    raise Exception("Deprecated - use get_coeffs_linreg instead")

    # coeffs = pd.Series(reg.coef_,
    #                    index= reg.feature_names_in_,
    #                    name=name)

    # if reg.intercept_ != 0.0:
    #     coeffs.loc['Intercept'] = reg.intercept_
    # return coeffs
    
def get_coeffs_linreg(lin_reg, feature_names=None, sort=True, ascending=True,
                      name='LinearRegression Coefficients'):
    """
    Get the coefficients of a linear regression model.

    Parameters:
    - lin_reg: The trained linear regression model.
    - feature_names: Optional. The names of the features used in the model. If not provided, it will use the feature names from the model.
    - sort: Optional. Whether to sort the coefficients by value. Default is True.
    - ascending: Optional. Whether to sort the coefficients in ascending order. Default is True.
    - name: Optional. The name of the coefficients series. Default is 'LinearRegression Coefficients'.

    Returns:
    - coeffs: A pandas Series containing the coefficients of the linear regression model.
    """
    if feature_names is None:
        feature_names = lin_reg.feature_names_in_
    ## Saving the coefficients
    coeffs = pd.Series(lin_reg.coef_, index=feature_names)
    coeffs['intercept'] = lin_reg.intercept_
    if sort == True:
        coeffs = coeffs.sort_values(ascending=ascending)
    return coeffs


def plot_coefficients(coeffs, figsize=(6,5), title='Regression Coefficients',
                      intercept=True, intercept_name='Intercept',
                      sort_values=True, ascending=True,
                      ):
    raise Exception("Deprecated: use plot_coeffs instead.")

#     ## Exclude intercept if intercept==False
#     if intercept==False:
#         if intercept_name in coeffs:
#             coeffs = coeffs.drop(intercept_name).copy()

#     ## Sort values
#     if sort_values:
#         ceoffs = coeffs.sort_values(ascending=ascending)

#     ## Plot
#     ax = ceoffs.plot(kind='barh',figsize=figsize)

#     ## Customize Viz
#     ax.axvline(0,color='k', lw=1)
#     ax.set(ylabel='Feature Name',xlabel='Coefficient',title=title)
#     return ax

def plot_coeffs(coeffs, top_n=None, figsize=(4,5), 
                intercept=False, intercept_name="intercept", 
                annotate=False, ha='left', va='center', size=12, 
                xytext=(4,0), textcoords='offset points'):
    """ Plots the top_n coefficients from a Series, with optional annotations.
    
    Parameters:
    coeffs (pd.Series): The coefficients to be plotted.
    top_n (int, optional): The number of top coefficients to plot. If None, all coefficients will be plotted. Default is None.
    figsize (tuple, optional): The size of the figure. Default is (4, 5).
    intercept (bool, optional): Whether to include the intercept coefficient in the plot. Default is False.
    intercept_name (str, optional): The name of the intercept coefficient. Default is "intercept".
    annotate (bool, optional): Whether to annotate the coefficients on the plot. Default is False.
    ha (str, optional): The horizontal alignment of the annotations. Default is 'left'.
    va (str, optional): The vertical alignment of the annotations. Default is 'center'.
    size (int, optional): The font size of the annotations. Default is 12.
    xytext (tuple, optional): The offset of the annotations from the data points. Default is (4, 0).
    textcoords (str, optional): The coordinate system used for the annotations. Default is 'offset points'.
    
    Returns:
    matplotlib.axes.Axes: The plot of the coefficients.
    """
    # Drop intercept if intercept=False and 
    if (intercept == False) & (intercept_name in coeffs.index):
        coeffs = coeffs.drop(intercept_name)
    if top_n == None:
        ## sort all features and set title
        plot_vals = coeffs.sort_values()
        title = "All Coefficients - Ranked by Magnitude"
    else:
        ## rank the coeffs and select the top_n
        coeff_rank = coeffs.abs().rank().sort_values(ascending=False)
        top_n_features = coeff_rank.head(top_n)
        
        ## sort features and keep top_n and set title
        plot_vals = coeffs.loc[top_n_features.index].sort_values()
        title = f"Top {top_n} Largest Coefficients"
    ## plotting top N importances
    ax = plot_vals.plot(kind='barh', figsize=figsize)
    ax.set(xlabel='Coefficient', 
            ylabel='Feature Names', 
            title=title)
    ax.axvline(0, color='k')
    if annotate == True:
        annotate_hbars(ax, ha=ha, va=va, size=size, xytext=xytext, textcoords=textcoords)
    return ax




def plot_residuals(model,X_test_df, y_test,figsize=(12,5)):
    """
    Plots a Q-Q Plot and residual plot for a statsmodels OLS regression.

    Parameters:
    model (statsmodels.regression.linear_model.RegressionResultsWrapper): The fitted regression model.
    X_test_df (pandas.DataFrame): The test dataset features.
    y_test (array-like): The test dataset target variable.
    figsize (tuple, optional): The size of the figure. Defaults to (12,5).

    Returns:
    None
    """
    ## Make predictions and calculate residuals
    y_pred = model.predict(X_test_df)
    resid = y_test - y_pred

    fig, axes = plt.subplots(ncols=2,figsize=figsize)

    ## Normality
    sm.graphics.qqplot(resid, line='45',fit=True,ax=axes[0]);

    ## Homoscedascity
    ax = axes[1]
    ax.scatter(y_pred, resid, edgecolor='white',lw=0.5)
    ax.axhline(0,zorder=0)
    ax.set(ylabel='Residuals',xlabel='Predicted Value');
    fig.tight_layout()
    plt.show()



def summarize_df(df_):
    """
    Summarizes a DataFrame by providing insights on column data types, null values, unique values, and numeric range.

    Parameters:
    df_ (pandas.DataFrame): The DataFrame to be summarized.

    Returns:
    pandas.DataFrame: A summary report DataFrame with the following columns:
        - 'Column': The column names of the DataFrame.
        - 'dtype': The data types of the columns.
        - '# null': The number of null values in each column.
        - 'null (%)': The percentage of null values in each column.
        - 'nunique': The number of unique values in each column.
        - 'min': The minimum numeric value in each column.
        - 'max': The maximum numeric value in each column.

    Example Usage:
    >> df = pd.read_csv(filename)
    >> summary = summarize_df(df)
    """
    df = df_.copy()
    report = pd.DataFrame({
                        'dtype':df.dtypes,
                        '# null': df.isna().sum(),
                        'null (%)': df.isna().sum()/len(df)*100,
                        'nunique':df.nunique(),
                        "min":df.min(numeric_only=True),
                        'max':df.max(numeric_only=True)
             })
    report.index.name='Column'

    with pd.option_context("display.min_rows", len(df)):
        display(report.round(2))
    return report.reset_index()


def get_importances(model, feature_names=None, name='Feature Importance',
                   sort=False, ascending=True):
    """
    Extract the feature importances for a given model.

    Parameters:
    model (object): The trained model for which feature importances are calculated.
    feature_names (list, optional): List of feature names. If not provided, it will be extracted from the model.
    name (str, optional): Name of the feature importances series. Default is 'Feature Importance'.
    sort (bool, optional): Whether to sort the importances in ascending order. Default is False.
    ascending (bool, optional): Whether to sort the importances in ascending order. Default is True.

    Returns:
    importances (pd.Series): Series containing the feature importances.
    """
    ## checking for feature names
    if feature_names == None:
        feature_names = model.feature_names_in_

    ## Saving the feature importances
    importances = pd.Series(model.feature_importances_, index= feature_names,
                           name=name)

    # sort importances
    if sort == True:
        importances = importances.sort_values(ascending=ascending)

    return importances




def plot_importance(importances, top_n=None,  figsize=(8,6)):
    """
    Plots the importance of features in a horizontal bar chart.

    Parameters:
    importances (pandas.Series): The importance values of the features.
    top_n (int, optional): The number of top most important features to plot. If None, all features will be plotted. Default is None.
    figsize (tuple, optional): The size of the figure. Default is (8, 6).

    Returns:
    matplotlib.axes.Axes: The axes object of the plot.

    """
    # sorting with asc=false for correct order of bars
    if top_n==None:
        ## sort all features and set title
        plot_vals = importances.sort_values()
        title = "All Features - Ranked by Importance"
    else:
        ## sort features and keep top_n and set title
        plot_vals = importances.sort_values().tail(top_n)
        title = f"Top {top_n} Most Important Features"
    ## plotting top N importances
    ax = plot_vals.plot(kind='barh', figsize=figsize)
    ax.set(xlabel='Importance',
            ylabel='Feature Names',
            title=title)
    ## return ax in case want to continue to update/modify figure
    return ax



def get_color_dict(importances, color_rest='#006ba4' , color_top='green',
                    top_n=7):
    """
    Returns a dictionary mapping feature names to colors based on their importances.

    Parameters:
    importances (pd.Series): A pandas Series containing feature importances.
    color_rest (str, optional): The color code for non-highlighted features. Defaults to '#006ba4'.
    color_top (str, optional): The color code for highlighted features. Defaults to 'green'.
    top_n (int, optional): The number of top features to highlight. Defaults to 7.

    Returns:
    dict: A dictionary mapping feature names to colors.
    """
    highlight_feats = importances.sort_values(ascending=True).tail(top_n).index
    colors_dict = {col: color_top if col in highlight_feats else color_rest for col in importances.index}
    return colors_dict


def plot_importance_color(importances, top_n=None,  figsize=(8,6), 
                          color_dict=None, ax=None):
    """
    Plot the feature importances with optional color highlighting.

    Parameters:
    - importances (pandas.Series): The feature importances.
    - top_n (int, optional): The number of top features to display. If None, all features will be displayed. Default is None.
    - figsize (tuple, optional): The figure size. Default is (8, 6).
    - color_dict (dict, optional): A dictionary mapping feature names to colors for highlighting. Default is None.
    - ax (matplotlib.axes.Axes, optional): The axes object to plot on. If None, a new figure and axes will be created. Default is None.

    Returns:
    - ax (matplotlib.axes.Axes): The axes object containing the plot.

    Example Use:
    fig, axes = plt.subplots(ncols=2, figsize=(20,8))
    n = 20 # setting the # of features to use for both subplots
    
    plot_importance_color(importances, top_n=n, ax=axes[0], color_dict= colors_top7)
    axes[0].set(title='R.F. Importances')

    plot_importance_color(permutation_importances, top_n=n, ax=axes[1], color_dict=colors_top7)
    axes[1].set(title='Permutation Importances')
    fig.tight_layout()
    """
    # sorting with asc=false for correct order of bars
    if top_n==None:
        ## sort all features and set title
        plot_vals = importances.sort_values()
        title = "All Features - Ranked by Importance"
    else:
        ## sort features and keep top_n and set title
        plot_vals = importances.sort_values().tail(top_n)
        title = f"Top {top_n} Most Important Features"
    ## create plot with colors, if provided
    if color_dict is not None:
        ## Getting color list and saving to plot_kws
        colors = plot_vals.index.map(color_dict)
        ax = plot_vals.plot(kind='barh', figsize=figsize, color=colors, ax=ax)
        
    else:
        ## create plot without colors, if not provided
        ax = plot_vals.plot(kind='barh', figsize=figsize, ax=ax)
        
    # set titles and axis labels
    ax.set(xlabel='Importance', 
           ylabel='Feature Names', 
           title=title)
    
    ## return ax in case want to continue to update/modify figure
    return ax


# def get_coeffs_logreg(logreg, feature_names = None, sort=True,ascending=True,
#                       name='LogReg Coefficients', class_index=0):
#     if feature_names is None:
#         feature_names = logreg.feature_names_in_ 
    
#     ## Saving the coefficients
#     coeffs = pd.Series(logreg.coef_[class_index],
#                        index= feature_names, name=name)
    
#     # use .loc to add the intercept to the series
#     coeffs.loc['intercept'] = logreg.intercept_[class_index]
#     if sort == True:
#         coeffs = coeffs.sort_values(ascending=ascending)  
#     return coeffs



def get_coeffs_logreg(logreg, feature_names = None, sort=True,ascending=True,
                      name='LogReg Coefficients', class_index=0,  
                      include_intercept=True, as_odds=False):
    """
    Get the coefficients of a logistic regression model.

    Parameters:
    logreg (object): The logistic regression model.
    feature_names (list, optional): List of feature names. If None, it uses the feature names from the model.
    sort (bool, optional): Whether to sort the coefficients. Default is True.
    ascending (bool, optional): Whether to sort the coefficients in ascending order. Default is True.
    name (str, optional): Name of the coefficients. Default is 'LogReg Coefficients'.
    class_index (int, optional): Index of the class for which to get the coefficients. Default is 0.
    include_intercept (bool, optional): Whether to include the intercept in the coefficients. Default is True.
    as_odds (bool, optional): Whether to exponentiate the coefficients to obtain odds ratios. Default is False.

    Returns:
    pd.Series: Series containing the coefficients.
    """
    
    if feature_names is None:
        feature_names = logreg.feature_names_in_
        
    ## Saving the coefficients
    coeffs = pd.Series(logreg.coef_[class_index],
                       index= feature_names, name=name)
    
    if include_intercept:
        # use .loc to add the intercept to the series
        coeffs.loc['intercept'] = logreg.intercept_[class_index]
        
    if as_odds==True:
        coeffs = np.exp(coeffs)
    if sort == True:
        coeffs = coeffs.sort_values(ascending=ascending)
    
        
    return coeffs





def plot_coeffs_color(coeffs, top_n=None,  figsize=(8,6), legend_loc='best',
                      threshold=None, color_lt='darkred', color_gt='forestgreen',
                      color_else='gray', label_thresh='Equally Likely',
                      label_gt='More Likely', label_lt='Less Likely',
                   plot_kws = {}):
    """Plots series of coefficients
    
    Args:
        coeffs (pandas Series): Importance values to plot.
        top_n (int): The number of features to display (Default=None).
                     If None, display all. Otherwise, display top_n most important.
        figsize (tuple): figsize tuple for .plot.
        legend_loc (str): Location of the legend in the plot (Default='best').
        threshold (float): Threshold value for coloring the coefficients (Default=None).
        color_lt (str): Color for coefficients less than the threshold (Default='darkred').
        color_gt (str): Color for coefficients greater than the threshold (Default='forestgreen').
        color_else (str): Color for coefficients that do not meet the threshold (Default='gray').
        label_thresh (str): Label for the threshold line in the legend (Default='Equally Likely').
        label_gt (str): Label for coefficients greater than the threshold in the legend (Default='More Likely').
        label_lt (str): Label for coefficients less than the threshold in the legend (Default='Less Likely').
        plot_kws (dict): Additional keyword arguments accepted by pandas' .plot method.
        
    Returns:
        matplotlib.axes._subplots.AxesSubplot: Matplotlib axis object.
    """
    # sorting with asc=false for correct order of bars
    if top_n is None:
        ## sort all features and set title
        plot_vals = coeffs.sort_values()
        title = "All Coefficients"
    else:
        ## rank the coeffs and select the top_n
        coeff_rank = coeffs.abs().rank().sort_values(ascending=False)
        top_n_features = coeff_rank.head(top_n)
        plot_vals = coeffs.loc[top_n_features.index].sort_values()
        ## sort features and keep top_n and set title
        title = f"Top {top_n} Largest Coefficients"
        ## plotting top N importances
    if threshold is not None:
        color_dict = get_colors_gt_lt(plot_vals, threshold=threshold,
                                      color_gt=color_gt,color_lt=color_lt,
                                      color_else=color_else)
        ## Getting color list and saving to plot_kws
        colors = plot_vals.index.map(color_dict)
        plot_kws.update({'color':colors})
    
    
    ax = plot_vals.plot(kind='barh', figsize=figsize,**plot_kws)
    ax.set(xlabel='Coefficient',
            ylabel='Feature Names',
            title=title)
    
    if threshold is not None:
        ln1 = ax.axvline(threshold,ls=':',color='black')
        from matplotlib.patches import Patch
        box_lt = Patch(color=color_lt)
        box_gt = Patch(color=color_gt)
        handles = [ln1,box_gt,box_lt]
        labels = [label_thresh,label_gt,label_lt]
        ax.legend(handles,labels, loc=legend_loc)
    ## return ax in case want to continue to update/modify figure
    return ax



def get_colors_gt_lt(coeffs, threshold=1, color_lt ='darkred',
                     color_gt='forestgreen', color_else='gray'):
    """
    Creates a dictionary of features and their corresponding colors based on whether the value is greater than or less than the threshold.

    Parameters:
    coeffs (pandas.DataFrame): The coefficients dataframe.
    threshold (float): The threshold value. Default is 1.
    color_lt (str): The color for values less than the threshold. Default is 'darkred'.
    color_gt (str): The color for values greater than the threshold. Default is 'forestgreen'.
    color_else (str): The color for values equal to the threshold. Default is 'gray'.

    Returns:
    dict: A dictionary mapping features to their respective colors.
    """
    colors_dict = {}
    for i in coeffs.index:
        rounded_coeff = np.round( coeffs.loc[i],3)
        if rounded_coeff < threshold:
            color = color_lt
        elif rounded_coeff > threshold:
            color = color_gt
        else:
            color=color_else
        colors_dict[i] = color
    return colors_dict





#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

import setuptools

# Function to load requirements from a file
def load_requirements(filename):
    try:
        with open(filename, 'r') as file:
            requirements = []
            for line in file:
                # Remove whitespace and skip comments and empty lines
                line = line.strip()
                if line and not line.startswith('#'):
                    requirements.append(line)
            return requirements
    except FileNotFoundError:
        print(f"Error: The file {filename} does not exist.")
        return []
    except Exception as e:
        print(f"An error occurred: {e}")
        return []


# ## Grab requirements form file
# with open("./requirements.txt") as f:
#     req_list = f.readlines()
#     req_list = [x.strip() for x in req_list if not x.startswith('#') and x.strip() != '']
#     # req_LIST = [str(req) for req in req_list]
# # req_list = ['scikit-learn>=1.2.2','pandas>=1.5.3','matplotlib>=3.7.1', 'seaborn>=0.12.2'
# #             'statsmodels>=0.13.5']#, 'plotly>=5.15.0']
# requirements = ['Click>=7.0', *req_list  ]

test_requirements = ['sphinx_rtd_theme' ]

setup(
    author="James Irving",
    author_email='jirving@codingdojo.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        
    ],
    description="Code from Coding Dojo's Online Part-Time Data Science boot camp",
    entry_points={
        'console_scripts': [
            'dojo_ds=dojo_ds.cli:main',
        ],
    },
    install_requires=load_requirements("./requirements.txt"),
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='dojo_ds',
    name='dojo_ds',
    packages=find_packages(include=['dojo_ds', 'dojo_ds.*']),
    test_suite='tests',
    tests_require=load_requirements("./requirements.txt") + test_requirements,
    url='https://github.com/coding-dojo-data-science/dojo_ds',
    version='1.1.2',
    zip_safe=False,
)



from .evaluate import evaluate_ols, plot_residuals
def find_outliers_Z(data, verbose=True):
    """
    Find outliers based on Z-score rule (outliers have an absolute z-score that is >3)

    Args:
        data (Series): Pandas Series
        verbose (bool, optional): Print summary info about outliers. Defaults to True.

    Returns:
        Series: Boolean index for input data, where True = Outlier
    """
    import pandas as pd
    import numpy as np
    import scipy.stats as stats
    outliers = np.abs(stats.zscore(data))>3
    
    
    if verbose:
        n = len(outliers)
        print(f"- {outliers.sum():,} outliers found in {data.name} out of {n:,} rows ({outliers.sum()/n*100:.2f}%) using Z-scores.")

    outliers = pd.Series(outliers, index=data.index, name=data.name)
    return outliers


def find_outliers_IQR(data, verbose=True):
    """
    Find outliers based on IQR-rule (outliers are either 1.5 x IQR below 25% quantile and 1.5xIQR above 75% quantile).
    
    Args:
        data (Series): Pandas Series containing the data.
        verbose (bool, optional): If True, print summary information about outliers. Defaults to True.
    
    Returns:
        Series: Boolean index for input data, where True indicates an outlier.
    """
    import pandas as pd
    import numpy as np
    
    # Calculate q1 and q3 quantiles
    q3 = np.quantile(data, .75)
    q1 = np.quantile(data, .25)
    
    # Calculate IQR 
    IQR = q3 - q1
    
    # Set thresholds more than 1.5x IQR above Q3/below Q1
    upper_threshold = q3 + 1.5 * IQR
    lower_threshold = q1 - 1.5 * IQR
    
    # Identify outliers
    outliers = (data < lower_threshold) | (data > upper_threshold)
    
    if verbose:
        n = len(outliers)
        print(f"- {outliers.sum():,} outliers found in {data.name} out of {n:,} rows ({outliers.sum() / n * 100:.2f}%) using IQR.")
    
    outliers = pd.Series(outliers, index=data.index, name=data.name)
    return outliers




def remove_outliers(df_, method='iqr', subset=None, verbose=2):
    """Returns a copy of the input dataframe with outliers removed from selected columns using the specified method.

    Args:
        df_ (DataFrame): The input dataframe to copy and remove outliers from.
        method (str): The method of outlier removal. Options are 'iqr' or 'z'/'zscore'. Default is 'iqr'.
        subset (list or None): A list of column names to remove outliers from. If None, all numeric columns are used. Default is None.
        verbose (bool, int): If verbose==1, print only the overall summary. If verbose==2, print the detailed summary. Default is 2.

    Returns:
        DataFrame: A copy of the input dataframe with outliers removed.

    Raises:
        Exception: If the method is not 'iqr' or 'z'.

    Examples:
        >>> df = pd.DataFrame({'A': [1, 2, 3, 4, 5], 'B': [10, 20, 30, 40, 50]})
        >>> remove_outliers(df, method='iqr', subset=['A'], verbose=2)
        Returns a dataframe with outliers removed from column 'A' using the IQR rule.
    """
    import pandas as pd
    ## Make a copy of the input dataframe  
    df = df_.copy()
    
    ## Set verbose_func for calls to outlier funcs
    if verbose == 2:
        verbose_func = True
    else:
        verbose_func = False
        
    ## Set outlier removal function and name
    if method.lower() == 'iqr':
        find_outlier_func = find_outliers_IQR
        method_name = "IQR rule"
    elif 'z' in method.lower():
        find_outlier_func = find_outliers_Z
        method_name = 'Z-score rule'
    else:
        raise Exception('[!] Method must be either "iqr" or "z".')
        
    ## Set list of columns to remove outliers from
    if subset is None:
        col_list = df.select_dtypes('number').columns
    elif isinstance(subset, str):
        col_list = [subset]
    elif isinstance(subset, list):
        col_list = subset
    else:
        raise Exception("[!] subset must be None, a single string, or a list of strings.")

    
    ## Empty dictionary for both types of outliers
    outliers = {}

    ## Use both functions to see the comparison for the number of outliers
    for col in col_list:
        idx_outliers = find_outlier_func(df[col], verbose=verbose_func)
        outliers[col] = idx_outliers

    
    ## Getting final dataframe of all outliers to get 1 final T/F index
    outliers_combined = pd.DataFrame(outliers).any(axis=1)
    
    if verbose:
        n = len(outliers_combined)
        print(f"\n[i] Overall, {outliers_combined.sum():,} rows out of {n:,} ({outliers_combined.sum()/n*100:.2f}%) were removed as outliers using {method_name}.")
    
    
    # remove outliers 
    df_clean = df[~outliers_combined].copy()
    return df_clean
      


import matplotlib.pyplot as plt
import seaborn as sns
fav_style = ('ggplot','tableau-colorblind10')
fav_context  ={'context':'notebook', 'font_scale':1.1}
plt.style.use(fav_style)
sns.set_context(**fav_context)
plt.rcParams['savefig.transparent'] = False
plt.rcParams['savefig.bbox'] = 'tight'